/* Statistical labeller - revised version.

   07-12-92 Created
   24-12-92 Restructured
   01-04-93 Change from using FSMs to parser
   14-04-93 FSMs reinstated
   05-04-93 Tag inference option added
   27-01-95 Unknown word handler added
   22-11-95 Minor fix in options checking
   16 Jan 2004 Cleaned up for open source release.
    
   Copyright (C) David Elworthy 2004.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   For details, see the file LICENCE, or visit http://www.gnu.org/copyleft/gpl.html

   Usage:
    label corpus options
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include "common.h"
#include "diction.h"
#include "trans.h"
#include "label.h"
#include "map.h"
#include "low.h"
#ifdef Analyse
#include "analyse.h"
#endif
#ifdef Use_Parser
#include "parser.h"
#endif
#ifdef Use_FSM
#include "fsm.h"
#endif

#define DICTLEN (60000)

/* Initialisation codes (input option 'I') */
#define Init_d_ntag   (1)
#define	Init_d_tagmax (2)
#define	Init_t_tagmax (4)
#define	Init_d_1      (8)
#define	Init_t_1      (16)
#define Init_d_s      (32)

/*-----------------------------------------------------------------------------
    check_names

    check dict and tran names have not already been specified.
-----------------------------------------------------------------------------*/

static BOOL check_names(char *dict, char *tran, char *kind)
{
    BOOL ok = TRUE;

    if (dict != NULL && dict[0] != 0)
    {
        fprintf(stderr, "%s dictionary specified more than once\n", kind);
        ok = FALSE;
    }

    if (tran != NULL && tran[0] != 0)
    {
        fprintf(stderr, "%s transitions specified more than once\n", kind);
        ok = FALSE;
    }

    return ok;
}

/*-----------------------------------------------------------------------------
    init_trans

    Initialise the trans and pi (but not gamma) arrays.
-----------------------------------------------------------------------------*/

static void init_trans(Trans *trans, Score s, BOOL mul)
{
    int size = trans->size;

    set_trans(trans->trans, s, size, mul);
    set_pi(trans->pi, s, size, mul);
}

/*-----------------------------------------------------------------------------
    set_up_options

    Set all the option flags, file names and numerical parameters.
-----------------------------------------------------------------------------*/

static void set_up_options(int argc, char *argv[],
    int *iterations, int *initialise, int *dict_size,
    char *dict, char *tran, char *odict, char *otran,
    char *out, char *map, char *skip, char *reduce,
    char *fsm, char *grammar, char *infer, char *ukw)
{
    char opt;
    int  arg = 2, i = 0;
    char root[MAXFN];
    BOOL error = FALSE;

    /* The following string defines the known option letters. ':' indicates
       that an argument is needed. */
    char *opt_string =
        "aA:bB:c:C:d:D:e:EfFgGi:I:lL:m:M:nNo:O:pPq:Qr:R:St:T:uUVwx:XzZ";

    *iterations = 1;
    *initialise = 0;
    *dict_size  = 0;
    SetOutOpt(out_word);
    SetInOpt(tagged);
    dict[0] = tran[0] = odict[0] = otran[0] = out[0] = map[0] =
    fsm[0] = grammar[0] = infer[0] = skip[0] = reduce[0] = 0;

    /* Analyse the options */
    while ((opt = get_option(argc, argv, opt_string, &arg, &i)) != 0)
    {
        switch (opt)
        {
            case 'a': SetOption(anchored); break;
            case 'A':
                if (!get_opt_int(argv, dict_size, &arg, &i, opt))
                    *dict_size = 0;
                break;
            case 'b': SetOption(num_stabilise); break;
            case 'B':
                SetOption(reestimate);
                if (!get_opt_int(argv, iterations, &arg, &i, opt))
                    *iterations = 1;
                break;
            case 'c':
            {
                Score threshold;

                if (get_opt_double(argv, &threshold, &arg, &i, opt))
                {
                    SetOption(use_threshold);
                    set_output_threshold(threshold);
                }
                break;
            }
            case 'C':
                if (!get_opt_int(argv, &options.in, &arg, &i, opt))
                    options.in = tagged;
                break;
            case 'd':
                if (check_names(dict, NULL, "Input"))
                    get_opt_string(argv, dict, MAXFN, &arg, &i, opt);
                break;
            case 'D':
                if (check_names(odict, NULL, "Output"))
                    get_opt_string(argv, odict, MAXFN, &arg, &i, opt);
                break;
            case 'e':
                /* Have to get string so arg and i are straight */
                get_opt_string(argv, fsm, MAXFN, &arg, &i, opt);
#ifdef Use_FSM
                SetOption(use_fsm);
#else
                fprintf(stderr, "Option 'e' ignored (no FSMs)\n");
#endif
                break;
            case 'E':
#ifdef Use_FSM
                SetOption(fsm_trace);
#else
                fprintf(stderr, "Option 'E' ignored (no FSMs)\n");
#endif
                break;
            case 'f': SetOption(most_freq); break;
            case 'F': SetOption(fb_tagging); break;
            case 'g': SetOption(correction); break;
            case 'G': SetOption(correction_lex); break;
            case 'i':
                get_opt_string(argv, infer, MAXFN, &arg, &i, opt);
                break;
            case 'I':
                if (!get_opt_int(argv, initialise, &arg, &i, opt))
                    *initialise = 0;
                break;
            case 'l': SetOption(training); break;
            case 'L':
            {
                Score threshold;
                if (get_opt_double(argv, &threshold, &arg, &i, opt))
                {
                    SetOption(reest_threshold);
                    set_re_est_threshold(threshold);
                }
                break;
            }
            case 'm':
                if (map[0] != 0)
                    fprintf(stderr,
                        "Map file name specified more than once\n");
                else
                    get_opt_string(argv, map, MAXFN, &arg, &i, opt);
                break;
            case 'M':
                if (reduce[0] != 0)
                    fprintf(stderr,
                        "Reduced tag set file specified more than once\n");
                else
                {
                    get_opt_string(argv, reduce, MAXFN, &arg, &i, opt);
                    SetOption(reduced_tags);
                }
                break;
            case 'n': SetOption(any_digit); break;
            case 'N': SetOption(parsed_number); break;
            case 'o':
                if (out[0] != 0)
                    fprintf(stderr,
                        "Output file name specified more than once\n");
                else
                    get_opt_string(argv, out, MAXFN, &arg, &i, opt);
                break;
            case 'O':
                if (!get_opt_int(argv, &options.out, &arg, &i, opt))
                    options.out = out_word;
                if (!no_output) SetOutOpt(out_word);
                break;
            case 'p': SetOption(product); break;
            case 'P':
#ifdef Phrasal
                SetOption(anchor_bracket);
#else
                fprintf(stderr, "Option 'P' ignored (not phrasal)\n");
#endif
                break;
            case 'q':
                get_opt_string(argv, grammar, MAXFN, &arg, &i, opt);
#ifdef Use_Parser
                SetOption(use_parser);
#else
                fprintf(stderr, "Option 'q' ignored (no parser)\n");
#endif
                break;
            case 'Q':
#ifdef Use_Parser
                SetOption(parser_trace);
#else
                fprintf(stderr, "Option 'Q' ignored (no parser)\n");
#endif
                break;
            case 'S': SetOption(report_stats); break;
            case 'r':
                if (check_names(dict, tran, "Input") &&
                    get_opt_string(argv, root, MAXFN, &arg, &i, opt))
                    make_names(root, dict, tran, MAXFN);
                break;
            case 'R':
                if (check_names(odict, otran, "Output") &&
                    get_opt_string(argv, root, MAXFN, &arg, &i, opt))
                    make_names(root, odict, otran, MAXFN);
                break;
            case 't':
                if (check_names(NULL, tran, "Input"))
                    get_opt_string(argv, tran, MAXFN, &arg, &i, opt);
                break;
            case 'T':
                if (check_names(NULL, otran, "Output"))
                    get_opt_string(argv, otran, MAXFN, &arg, &i, opt);
                break;
            case 'u': SetOption(report_unknown); break;
            case 'U':
                if (get_opt_string(argv, ukw, MAXFN, &arg, &i, opt))
                    SetOption(unknown_rules);
                break;
            case 'V': SetOption(Viterbi); break;
            case 'w': SetOption(use_wordlist); break;
            case 'x':
                if (check_names(NULL, skip, "Skip list"))
                {
                    get_opt_string(argv, skip, MAXFN, &arg, &i, opt);
                    SetOption(skip_list);
                }
                break;
            case 'X': SetOption(special); break;
            case 'z': SetOption(verbose); break;
            case 'Z': SetOption(debug); break;
        }
    }

    /* Set up default for mapping */
    if (map[0] == 0) strcpy(map, "tags.map");

    /* Fiddle iterations if training */
    if (Option(training))
    {
        if (*iterations < 1) *iterations = 1;
        if (Option(Viterbi)) *iterations = 2;
        if (Option(fb_tagging) && !Option(reestimate)) *iterations = 2;
        if (*dict_size <= 0) *dict_size = DICTLEN;
    }

    /* If there were no other tagging options, set f-b */
    if (!Option(most_freq) && !Option(Viterbi))
        SetOption(fb_tagging);

    /* Verification of options */
    /* Check numbers */
    if (Option(any_digit) && Option(parsed_number))
    {
        fprintf(stderr, "Both number options specified\n");
        error = TRUE;
    }

    /* Check input files */
    if (Option(training) && Option(most_freq))
    {
        fprintf(stderr,
            "Training must be a separate run from 'most frequent' tagging\n");
        error = TRUE;
    }

    if (Option(training) && Option(unknown_rules))
    {
        fprintf(stderr,
            "Training must be a separate run from tagging with unknown word rules\n");
        error = TRUE;
    }

    if (!Option(training) && (Option(correction) || Option(correction_lex)))
    {
        fprintf(stderr, "Correction adjustment only applies when training\n");
        ClearOption( correction );
        ClearOption( correction_lex );
        /* Just a warning! */
    }

    if (dict[0] == 0 && (!Option(training) || Option(use_wordlist)))
    {
        fprintf(stderr, "Must specify input dictionary (unless training)\n");
        error = TRUE;
    }
    
    if (tran[0] == 0 && !Option(training) &&
        !(*initialise & ( Init_t_1 | Init_t_tagmax )) )
    {
        fprintf(stderr,
            "Must specify either transitions file or initialisation option\n");
        error = TRUE;
    }
    if (Option(training) && !Option(use_wordlist) &&
        (dict[0] != 0 || tran[0] != 0))
    {
        fprintf(stderr,
            "Dictionary/transitions input file ignored for training run\n");
        dict[0] = tran[0] = 0;
    }
    if (Option(use_wordlist) && dict[0] == 0)
    {
        fprintf(stderr, "Dictionary must be specified for wordlist\n");
        error = TRUE;
    }

    if (dict[0] == 0 && infer[0] != 0)
    {
        fprintf(stderr,
            "Tag inference may only be specified with an input dictionary\n");
        error = TRUE;
    }

    /* Check corpus */
    if (Option(training) && !Option(use_wordlist) && InOpt(untagged_input))
    {
        fprintf(stderr, "Training requires a tagged corpus\n");
        error = TRUE;
    }

    /* Check output files and options */
    if (Option(Viterbi) && (OutOpt(all_tags) || OutOpt(out_scores)))
    {
        fprintf(stderr, "Viterbi run: output options ignored\n");
        ClearOutOpt(all_tags | out_scores);
    }
    if (InOpt(untagged_input))
    {
        ClearOutOpt(err_only);
        ClearOption(report_stats);
        ClearOutOpt(prob_dist);
#ifdef Analyse
        ClearOutOpt(analyse);
#endif
    }
    if (Option(training) && (*iterations == 1) &&
        odict[0] == 0 && otran[0] == 0)
    {
        fprintf(stderr, "Warning: training run with no output files\n");
    }

    /* Check tagging options */
    if (Option(Viterbi) && (*iterations != (Option(training) ? 2 : 1)))
    {
        fprintf(stderr, "Iterations parameter ignored for Viterbi run\n");
        *iterations = 1;
    }
    if (*iterations < 1)
    {
        fprintf(stderr, "Number of iterations must be 1 or more\n");
        error = TRUE;
    }
    if (Option(training) && (*iterations == 1) && !no_output)
    {
        fprintf(stderr,
            "Warning: 'No output' set since training and 1 iteration\n");
        options.out = no_out_opt;
        ClearOption(report_stats);
        ClearOutOpt(prob_dist);
    }
    if (Option(most_freq) && (Option(Viterbi) || Option(reestimate)))
    {
        fprintf(stderr, "'Most frequent option set: ignoring others\n");
        ClearOption(Viterbi);
        ClearOption(reestimate);
    }

    if (Option(reest_threshold) && !Option(reestimate))
    {
        fprintf(stderr, "Re-estimation threshold ignored\n");
        ClearOption(reest_threshold);
    }
    
    if (OutOpt(prob_dist) &&
        ((Option(training) && *iterations == 1) || Option(most_freq) ||
            Option(Viterbi)))
    {
        fprintf(stderr, "Probability distribution option ignored\n");
        ClearOutOpt(prob_dist);
    }    

    if (Option(use_threshold) &&
        ((Option(training) && *iterations == 1) || Option(most_freq) ||
            Option(Viterbi)))
    {
        fprintf(stderr, "Thresholding distribution option ignored\n");
        ClearOption(use_threshold);
    }    

    /* Check initialisation options */
    if (Option(product) & !Option(training))
    {
        fprintf(stderr,
            "Product option ignored except when training\n");
        ClearOption(product);
    }

    /* Check we made it through unscathed */
    if (error) get_out();
}

/*-----------------------------------------------------------------------------
    main

    Entry point.
-----------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{                    
    int   iterations, initialise, iter, dict_size;
    options_st saved_options;
    char  dictname[MAXFN], tranname[MAXFN], odictname[MAXFN],
        otranname[MAXFN], outname[MAXFN], mapname[MAXFN],
        skipname[MAXFN], reducename[MAXFN], infername[MAXFN];
    char  fsmname[MAXFN], grammarname[MAXFN], ukwname[MAXFN];
    Dict  dict, skip_dict;
    Trans trans, newtrans;
    FILE  *corpfile;
    FILE  *outfile, *odictfile, *otranfile;
    BOOL  any_output;

    setlocale(LC_CTYPE, "iso_8858_1");
#ifdef SpecialMalloc
    /* Force fast allocation */
    set_small_allocation(100);
#endif

    /* Clear data structures */
    InitDict(dict)
    InitDict(skip_dict)
    InitTrans(trans)
    InitTrans(newtrans)
    odictfile = otranfile = NULL;

    /* Verify command line */
    if (argc <= 2)
        error_exit("Usage: label corpus options\n");

    /* Form options */
    InitOptions;
    
    set_up_options(argc, argv, &iterations, &initialise, &dict_size,
        dictname, tranname, odictname, otranname, outname, mapname,
        skipname, reducename, fsmname, grammarname, infername, ukwname);
    any_output = !no_output || Option(report_stats) || OutOpt(prob_dist);

    /* Read mappings */
    if (Option(verbose)) printf("Read mappings\n");
    read_mapping(mapname);

    /* Read tag reduction mappings */
    if (Option(reduced_tags))
    {
        if (Option(verbose)) printf("Read reduced tag set\n");
        read_reduce_mapping(reducename);
    }

#ifdef Use_Parser
    /* Read parse rules */
    if (Option(use_parser))
    {
        if (Option(verbose)) printf("Read parse rules\n");
        parser_read_named(grammarname);
    }
#endif
#ifdef Use_FSM
    /* Read FSM definitions */
    if (Option(use_fsm))
    {
        if (Option(verbose)) printf("Read FSMs\n");
        fsm_read_named(fsmname);
    }
#endif

    /* Read skip list */
    if (Option(skip_list))
    {
        if (Option(verbose)) printf("Read skip list\n");
        read_named_dict(skipname, &skip_dict, -1);
    }

    /* Read unknown word rules */
    if (Option(unknown_rules))
    {
        if (Option(verbose)) printf("Read unknown word rules\n");
        read_unknown_rules(ukwname);
    }

    /* Set up dictionary */
    if (dictname[0] == 0)
    {
        create_dict(&dict, dict_size);
        clear_dict(&dict);
    }
    else
    {
        if (Option(verbose)) printf("Read dictionary\n");
        read_named_dict(dictname, &dict, -1);
        if (infername[0] != 0)
        {
            if (Option(verbose)) printf("Read inference rules\n");
            infer_tags(infername, &dict);
        }
    }
    set_special_words(&dict);
    
    /* Set up transitions */
    if (tranname[0] == 0)
    {
        create_trans(&trans, tags_all);
        clear_trans_all(&trans);
    }
    else
    {
        if (Option(verbose)) printf("Read transitions\n");
        read_named_trans(tranname, &trans);
    }

    /* Create space for re-estimation or training */
    if (Option(reestimate) || Option(training))
    {
        newtrans.gamma = trans.gamma; /* Share arrays */
        create_trans(&newtrans, tags_all);
    }

    /* Open input corpus, '-' for stdin */
    if ( argv[1][0] == '-' ) corpfile = stdin;
    else corpfile = open_file(argv[1], "r");

    /* Check output files can be opened */
    if (any_output)
        outfile = (outname[0] == 0) ? stdout : open_file(outname, "w");
    if (odictname[0] != 0)
        odictfile = open_file(odictname, "w");
    if (otranname[0] != 0)
        otranfile = open_file(otranname, "w");

    /* Set up anchor word */
    set_anchor(&dict);

    /* -------- The main tagging loop -------- */
    /* Clear output options except for last iteration */
    saved_options = options;
    if (iterations > 1) options.out = no_out_opt;
    
    for (iter = 1 ; iter <= iterations ; iter++)
    {
        if (Option(verbose))
            printf("\nIteration %d%s", iter,
                Option(training) ? " (training)\n" : "\n");
        reset_corpus(corpfile);

#if (0)
        /* Report current space usage */
        space_report(stdout);
#endif

        /* If training, clear other tagging options */
        if (Option(training))
        {
            ClearOption(Viterbi);
            ClearOption(fb_tagging);
            ClearOption(most_freq);
            ClearOption(num_stabilise);
            if ( Option(correction) )
                init_correction_trans(&trans);
            else
                clear_trans_all(&newtrans);
        }

        /* On the first iteration, apply initialisation codes (if appropriate);
            otherwise do gamma adjustments. */
        if (!Option(training) && !Option(most_freq))
        {
            if (iter == 1)
            {
                /* Dict initialisation */
                Score mul, div;

                mul = (initialise & Init_d_1)    ? 1 : 0;
                div = (initialise & Init_d_ntag) ? 0 :
                    (initialise & Init_d_tagmax) ? tags_all :
                    (initialise & Init_d_s) ? -1 : 1;
                
                if (initialise & (Init_d_ntag | Init_d_tagmax | Init_d_s))
                {
                    set_up_scores(&dict, mul, div);
                }
                else
                {   /* Use gamma as divisor, maybe set to 1 first */
                    if (initialise & Init_d_1)
                        set_up_scores(&dict, 1, 1);
                    adjust_dict(&dict, trans.gamma, FALSE);
                }

                /* Trans initialisation */
                if (initialise & Init_t_tagmax)
                {
                    init_trans(&trans, 1.0/tags_all,
                        (initialise & Init_t_1) == 0);
                }
                else
                {
                    if (initialise & Init_t_1)
                        init_trans(&trans, 1.0, FALSE);
                    adjust_trans(&trans, NULL);
                }

                initialise = 0;
            }
            else /* Do normal gamma adjustment */
            {
                adjust_dict(&dict, trans.gamma, TRUE);
                adjust_trans(&trans, &newtrans);
            }
        }

        /* Clear re-estimation parameters */
        if (Option(reestimate)) clear_trans_all(&newtrans);

        /* Now actually label the corpus */
        init_statistics();
        tag_corpus(corpfile, outfile, &dict, &skip_dict, &trans, &newtrans);

        /* Give results, except on iter 1 of training */
        if (!Option(training))
        {
            if (Option(report_stats))
            {
                fprintf(outfile, "\n\nIteration %d\n", iter);
                report_results(outfile);
                fflush(outfile);
            }
        }

        /* If we were training, sort the dictionary */
        if (Option(training))
        {
            if (Option(verbose)) printf("\nSorting dictionary\n");
            sort_dict(&dict);

            /* If there are more iterations after the training ... */
            if (iterations > 1)
            {
                /* Restore options */
                options = saved_options;

                /* Mark that we are no longer training */
                ClearOption(training);
                saved_options = options;
                options.out   = no_out_opt;
            }
        }

        /* Restore global options on last loop */
        if (iter == iterations - 1)
            options = saved_options;
    }

#if (0)
    /* Report current space usage */
    space_report(stdout);
#endif

    /* Call the results output routines */
#ifdef Analyse
    if (OutOpt(analyse))
        print_analysis(outfile);
#endif
    if (any_output && outfile != stdout)
        fclose(outfile);
    fclose(corpfile);

    /* Write new arrays */
    if (odictfile != NULL)
    {
        if (Option(verbose))
            printf("Writing dictionary (%d entries)\n", dict.size);
        write_dict(odictfile, &dict, Option(training) || Option(reestimate));
        fclose(odictfile);
    }
    if (otranfile != NULL)
    {
        if (Option(verbose))
            printf("Writing transitions\n");
        if (Option(training) || Option(reestimate))
            write_trans(otranfile, &newtrans);
        else
            write_trans(otranfile, &trans);
        fclose(otranfile);
    }

    return 0;
}
