/* Statistical labeller - stack management.

   04-01-93 Created
   28-01-93 Restructured stack
   09-02-93 Further substantial change to data structures
   24-03-93 Changes for phrases in input
   30-03-93 Restructuring
   01-04-93 Change from using FSMs to parser
   14-04-93 FSMs reinstated
   21-04-93 Move towards chart data structure
   04-05-93 Output functions moved to io.c
   27-01-95 Unknown word handler added
   30-06-95 Allow biasing of scores on unknown words
   16 Jan 2004 Cleaned up for open source release.
    
   Copyright (C) David Elworthy 2004.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   For details, see the file LICENCE, or visit http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "map.h"
#include "diction.h"
#include "label.h"
#include "stack.h"
#ifdef Phrasal
#include "phrase.h"
#endif
#ifdef Use_Parser
#include "parser.h"
#endif
#ifdef Use_FSM
#include "fsm.h"
#endif
#include "list.h"

/*
==============================================================================
Stack management
*/

/*-----------------------------------------------------------------------------
    nodes_dump

    Dumps a sequence of nodes in textual form, optionally pausing after
    output. If full is TRUE, gives addresses as well. 
-----------------------------------------------------------------------------*/

/* Indent a line */
void indent(FILE *out, int depth)
{
    int i;
    for (i = 0 ; i < depth ; i++) fputc(' ', out);
}

void nodes_dump(FILE *out, Node from, Node to, BOOL pause, BOOL full,int depth)
{
    Node n;

    /* Work through each node */
    for (n = from ; n != NULL ; n = Succ(n, to))
    {
        Link l;
        SHyp h;
        SHyp chosen = n->chosen;

        indent(out, depth);
        if (full)
            fprintf(out, "** Node %x (pred %x succ %x chosen %x)\n",
                (int)n, (int)(n->pred), (int)(n->succ),
                (int)(n->chosen));
        else
            fprintf(out, "** Node %x\n", (int)n);

        /* List all the base hypotheses associated with this node */
        indent(out, depth);
        if (n->base == NULL)
        {
            fprintf(out, "Base: null\n");
        }
        else
        {
            fprintf(out, "Base:\n");
            for (l = n->base ; l != NULL ; l = l->next)
            {
                Hyp hyp = l->u.hyp;

                indent(out, depth);
                if (full)
                    fprintf(out, "hyp %x: ", (int)hyp);

                switch (hyp->type)
                {
                    case LexHyp:
                        {
                            Lexeme l = hyp->lex_start;
                            fprintf(out, "(%d) %s_%s:%g", l->id, LexemeText(l),
                                unmap_tag(hyp->tag), hyp->score);
                            if (l != hyp->lex_end)
                                fprintf(out, " *!* lexeme inconsistency *!* ");
                            break;
                        }
#ifdef Phrasal
                    case PhraseHyp:
                        dump_phrase_hyp(out, hyp);
                        break;
#endif
#ifdef Use_Parser
                    case ActiveHyp:
                        parser_dump_edge(out, hyp);
                        break;
#endif
                }
                fprintf(out, "\n");
            }
        }

        /* Dump all scored hypotheses starting here */
        indent(out, depth);
        if (n->start == NULL)
        {
            fprintf(out, "Start: null\n");
        }
        else
        {
            fprintf(out, "Start:\n");
            forHypothesesFwd(n, h, l)
            {
                Hyp hyp = h->hyp;

                indent(out, depth);
                if (full)
                {
                    fprintf(out, "sh %x h %x start %x end %x%s ",
                        (int)h, (int)hyp,
                        (int)(h->start), (int)(h->end),
                        (h == chosen) ? " [chosen]" : "");
                }
                fprintf(out, "%s ", unmap_tag(hyp->tag));

                /* Print scores information */
                if (Option(fb_tagging))
                    fprintf(out, "a %g b %g g %g",
                        h->u.fb.alpha, h->u.fb.beta, h->u.fb.gamma);
                else if (Option(Viterbi))
                    fprintf(out, "s %g", h->u.v.s);
                
                fprintf(out, "\n");

#ifdef Phrasal
                /* For phrasal hypotheses, now recurse */
                if (hyp->type == PhraseHyp)
                    nodes_dump(out, hyp->p.phrase.start, hyp->p.phrase.end,
                        FALSE, full, depth+1);
#endif
            }
        }

        /* List all scored hypotheses ending here */
        if (full)
        {
            indent(out, depth);
            if (n->end == NULL)
            {
                fprintf(out, "End: null\n");
            }
            else
            {
                fprintf(out, "End:\n");
                forHypothesesBack(n, h, l)
                {
                    Hyp hyp = h->hyp;

                    indent(out, depth);
                    if (full)
                        fprintf(out, "sh %x h %x ", (int)h, (int)hyp);
                    fprintf(out, "%s\n", unmap_tag(hyp->tag));
                }
            }
        }

        /* Dump active states from this node */
#ifdef Use_FSM
        if (n->fsm_state != NULL)
        {
            indent(out, depth);
            fprintf(out, "(%d) Active states:\n", n->lex->id);
            fsm_dump_state(out, depth+1, n->fsm_state);
        }
#endif
    }
    fflush(stdout);

    if (pause)
    {
        char c;
        fprintf(out, "Continue? (y/n)\n");
        fflush(out);
        
        do
        {
            scanf("%c", &c);
            if (c == 'n') get_out();
        } while (c != 'y');
    }

    if (depth == 0) fprintf(out, "\n");
}

/*-----------------------------------------------------------------------------
    getlink

    Get a link from a list. If complete is set, skip active edges. If the link
    we are at is unacceptable, step forward until we find a good one. 
-----------------------------------------------------------------------------*/

static Link getlink(Link link, BOOL complete)
{
    do
    {
#ifdef Use_Parser
        /* Get out on NULL link or a good link */
        if (link == NULL || TypeOf(link->u.shyp) != ActiveHyp || !complete)
            return link;
#else
        /* Always return */
        return link;
#endif
    } while ((link = link->next) != NULL);

   return NULL;
}

/*-----------------------------------------------------------------------------
    nextlink

    Advance to the next link. If complete is set, skip active edges.
-----------------------------------------------------------------------------*/

Link nextlink(Link link, BOOL complete)
{
    return getlink((link == NULL) ? NULL : link->next, complete);
}

/*-----------------------------------------------------------------------------
    firstlink

    First first acceptable link.
-----------------------------------------------------------------------------*/

Link firstlink(Node node, BOOL start, BOOL complete)
{
    return getlink((start) ? node->start : node->end, complete);
}

/*-----------------------------------------------------------------------------
    create_hyp

    Create a hypothesis structure of the given type, tag and score. The
    lexemes are set into the fields which indicate the source of the
    hypothesis.
-----------------------------------------------------------------------------*/

Hyp create_hyp(Tag tag, Score s, HypType type, Lexeme l_start, Lexeme l_end)
{
    Hyp hyp;

    Allocate(hyp, sizeof(HypSt), "hypothesis");
    hyp->type   = type;
    hyp->tag    = tag;
    hyp->score  = s;
    hyp->lex_start = l_start;
    hyp->lex_end   = l_end;

    if (type == LexHyp)
        hyp->p.lex.d = NULL;
#ifdef Phrasal
    else if (type == PhraseHyp)
        hyp->p.phrase.start = hyp->p.phrase.end = NULL;
#endif
#ifdef Use_Parser
    else if (type == ActiveHyp)
        parser_init_hyp(hyp);
#endif

    return hyp;
}

/*-----------------------------------------------------------------------------
    link_to_base

    Place an unscored hypothesis in the base list of a node.
-----------------------------------------------------------------------------*/

void link_to_base(Hyp hyp, Node node)
{
    node->base = create_link(node->base, FALSE, hyp, NULL, NULL);
}

/*-----------------------------------------------------------------------------
    create_lex_hyp

    Create a lexical hypothesis structure with a pointer back to a dictionary
    entry. The hypotheses are not placed in a base list for a node.
-----------------------------------------------------------------------------*/

Hyp create_lex_hyp(Tag tag, Score s, TagScore d, Lexeme lex)
{
    Hyp hyp = create_hyp(tag, s, LexHyp, lex, lex);
    hyp->p.lex.d = d;
    return hyp;
}

/*-----------------------------------------------------------------------------
    create_shyp

    Create a scored hypothesis from a hypothesis.
-----------------------------------------------------------------------------*/

SHyp create_shyp(Hyp hyp, Node start_parent, Node end_parent)
{
    SHyp shyp;

    Allocate(shyp, sizeof(SHypSt), "shyp");
    shyp->hyp   = hyp;
    shyp->start = start_parent;
    shyp->end   = end_parent;

    /* Fill in fields specific to a labelling algorithm */
    if (Option(Viterbi))
    {
        shyp->u.v.s    = 0;
        shyp->u.v.prev = NULL;
    }
    else if (Option(fb_tagging))
    {
        shyp->u.fb.alpha = shyp->u.fb.beta = shyp->u.fb.gamma = 0;
    }

    return shyp;
}

/*-----------------------------------------------------------------------------
    create_link

    Create a link with a given hypothesis or scored hypothesis and link to
    give place. One of hyp and shyp should be NULL, depending on "scored".
-----------------------------------------------------------------------------*/

Link create_link(Link next, BOOL scored, Hyp hyp, SHyp shyp, Link *prev)
{
    Link link;

    Allocate(link, sizeof(LinkSt), "link");
    link->next = next;
    if (scored)
        link->u.shyp = shyp;
    else
        link->u.hyp  = hyp;

    if (prev != NULL) *prev = link;

    return link;
}

/*-----------------------------------------------------------------------------
    copy_hyps

    Copy the list of unscored hyps at base to a scored hyp list, and return
    the list of links so created. The parents of the shyp are set to the start
    and end nodes.
-----------------------------------------------------------------------------*/

Link copy_hyps(Link base, Node start, Node end)
{
    Link head, *prev = &head;

    for ( ; base != NULL ; base = base->next)
    {
        Link l = create_link(NULL, TRUE, NULL,
            create_shyp(base->u.hyp, start, end), prev);
        prev = &(l->next);
    }

    return head;   
}

/*-----------------------------------------------------------------------------
    copy_links

    Copy a scored links structure
-----------------------------------------------------------------------------*/

Link copy_links(Link from)
{
    Link head, *prev = &head;

    for ( ; from != NULL ; from = from->next)
    {
        Link l = create_link(NULL, TRUE, NULL, from->u.shyp, prev);
        prev = &(l->next);
    }

    return head;   
}

/*-----------------------------------------------------------------------------
    insert_link

    Insert a link or sequence of links into a list
-----------------------------------------------------------------------------*/
void insert_link(Link *prev, Link link)
{
    /* Find end */
    Link l, last;
    for (l = link ; l != NULL ; l = l->next) last = l;

    last->next = *prev;
    *prev = link;
}

/*-----------------------------------------------------------------------------
    copy_word

    Make a copy of a word structure.
-----------------------------------------------------------------------------*/

static Word copy_word(Word from)
{
    Word word;

    Allocate(word, sizeof(WordSt), "word");
    *word = *from;

    return word;
}

/*-----------------------------------------------------------------------------
    free_word

    Free a word structure.
-----------------------------------------------------------------------------*/

static void free_word(Word word)
{
    /* Free the text and left and right contexts */
    if (word->text  != NULL) free(word->text);
    if (word->left  != NULL) free(word->left);
    if (word->right != NULL) free(word->right);

    /* Free the word itself */
    free(word);
}

/*-----------------------------------------------------------------------------
    create_node

    Create a node and link it in.
-----------------------------------------------------------------------------*/

Node create_node(Node pred, Lexeme lex)
{
    Node new;

    Allocate(new, sizeof(NodeSt), "node");
    new->pred    = pred;
    new->succ    = NULL;
    new->base    = new->start = new->end = NULL;
    new->chosen  = NULL;
    new->lex     = lex;

    if (pred != NULL) pred->succ = new;

#ifdef Use_FSM
    fsm_init_node(new);
#endif
    return new;
}

/*-----------------------------------------------------------------------------
    create_lexeme

    Create a lexeme and a node, linking it to prev. Called on reading a new
    word.
-----------------------------------------------------------------------------*/

Lexeme create_lexeme(Lexeme prev, Word word, BOOL share_word)
{
    Lexeme new;
    static int id = 0;

    Allocate(new, sizeof(LexemeSt), "lexeme");
    new->word = (share_word) ? word : copy_word(word);
    new->node = create_node((prev == NULL) ? NULL : prev->node, new);
    new->id   = id++;

    return new;
}

/*-----------------------------------------------------------------------------
    copy_from_dict

    Create hypotheses from a dictword. Returns list of hypotheses, as a Link
    list with pointers to Hyps (not SHyps).
    If we are training, we actually copy from the correct tag.
-----------------------------------------------------------------------------*/

static Link copy_from_dict(Lexeme lp, DictWord d)
{
    /* Create a list of hypotheses */
    if (Option(training) && !Option(use_wordlist)) /* Copy correct tag */
    {
        Tag tag = lp->word->ctag;
        TagScore t;
        Link     l;

        /* Search for the dict entry with this tag */
        for (t = d->tag ; t != NULL ; t = t->next)
            if (tag == t->tag) break;

        l = create_link(NULL, FALSE, create_lex_hyp(tag, 1.0, t, lp),
            NULL, NULL);
        return l;
    }
    else /* Copy all from dictionary */
    {
        Link l, *last;
        Link head;
        TagScore t;

        last = &head;

        for (t = d->tag ; t ; t = t->next)
        {
            l = create_link(NULL, FALSE,
                create_lex_hyp(t->tag, t->score, t, lp),
                NULL, last);
            last = &(l->next);
        }

        return head;
    }
}

/*-----------------------------------------------------------------------------
    copy_unknown

    Create hypotheses from the unknown word entry.
    text is present in case we want to do some morphological analysis.
-----------------------------------------------------------------------------*/

static Link copy_unknown(Lexeme lp, uchar *text, DictWord d)
{
    Link word_link = copy_from_dict(lp, d);

    /* If we are using unknown rules, we now alter the tag list */
    if (Option(unknown_rules))
    {
        UnkRule rule = unknown_word_rules;
        while ( rule )
        {
            BOOL match = FALSE;  /* Fixed, 28-6-95: used to say true */
            if ( rule->class.prefix )
            {
                uchar *t = text;
                if ( rule->match_pattern )
                {
                    int len = strlen( rule->match_pattern );
                    match = ( strncmp( t, rule->match_pattern, len ) == 0 );
                    t += len;
                }

                if ( match )
                {
                    if ( rule->nomatch_pattern )
                    {
                        int nlen = strlen( rule->nomatch_pattern );
                        match = (strncmp(t, rule->nomatch_pattern, nlen) != 0);
                    }
                }
            }
            else if ( rule->class.suffix )
            {
                uchar *end = text + strlen(text);
                if ( rule->match_pattern )
                {
                    int len = strlen( rule->match_pattern );
                    end -= len;
                    match = ( strcmp( end, rule->match_pattern ) == 0 );
                }
                
                if ( match )
                {
                    if ( rule->nomatch_pattern )
                    {
                        int nlen = strlen( rule->nomatch_pattern );
                        match = (strncmp(end-nlen,rule->nomatch_pattern,nlen)
                            != 0);
                    }
                    else match = TRUE;
                }
            }
            else if ( rule->class.initial_cap )
            {
                /* Check first alpha character */
                int l;
                for ( l = 0 ; text[l] ; l++ )
                    if ( isalpha(text[l]) )
                    {
                        match = isupper( text[l] );
                        break;
                    }
            }
            else if ( rule->class.all_cap )
            {
                /* Check all alpha characters */
                int l;
                BOOL any_alpha = FALSE;
                for ( l = 0 ; text[l] != 0 ; l++ )
                    if ( isalpha(text[l]) )
                    {
                        any_alpha= TRUE;
                        if ( !isupper( text[l] ) ) break;
                    }
                if ( any_alpha && text[l] == 0 ) match = TRUE;
            }
            else if ( rule->class.any_cap )
            {
                /* Check all alpha characters */
                int l;
                for ( l = 0 ; text[l] != 0 ; l++ )
                    if ( isalpha(text[l]) && isupper( text[l] ) )
                    {
                        match = TRUE;
                        break;
                    }
            }

            if ( ( ( rule->class.negated_pattern == 1 ) && !match )
                || match )
            {
                Link link, *prev = &word_link, next;
                int  ntags = rule->ntags;

                /* Scan the tag list and unhook any we do not want */
                for ( link = word_link ; link != NULL ; link = next )
                {
                    Tag tag = link->u.hyp->tag;
                    int t;
                    next = link->next;

                    /* Check against each tag in the rule's list */
                    for ( t = 0 ; t < ntags ; t++ )
                    {
                        if ( tag == rule->tags[t] ) break;
                    }

                    /* Remove if not in list and not negated, or v-v */
                    if ( ( t == ntags && !(rule->class.negated_tags)) ||
                        ( t != ntags &&   rule->class.negated_tags ) )
                    {
                        *prev = next;
                        link->next = NULL;
                        free( link->u.hyp );
                        free( link );
                    }
                    else
                    {
                        /* Not removed */
                        /* If there is a score in the rule, bias with it */
                        if ( t != ntags && rule->scores[t] >= 0 )
                            link->u.hyp->score *= rule->scores[t];
                        prev = &( link->next );
                    }
                }

                /* If there is continuation, skip to it; otherwise, get out */
                if ( rule->cont != -1 )
                {
                    int cont = rule->cont;
                    while ( (rule = rule->next) != NULL
                        && rule->number != cont ) ;
                }
                else break;
            }
            else rule = rule->next;
        }

        /* If we ended up with a completely empty list, restore everything */
        if ( word_link == NULL )
        {
            fprintf( stderr, "Warning: all tags would be removed for '%s'\n",
                text );
            word_link = copy_from_dict(lp, d);
        }
    }

    return word_link;
}

/*-----------------------------------------------------------------------------
   push_word

   Pushes a word onto the stack. This creates all the hypotheses. prev points
   to the previous used level of the stack. Also sets up the succ and pred
   links, based on word class. Returns updated stack pointer.

   Also creates start and end lists which are the same as the base hypothesis
   list.
-----------------------------------------------------------------------------*/

Lexeme push_word(Word word, Lexeme prev)
{
    Lexeme new = create_lexeme(prev, word, FALSE);
    Node  node = new->node;

    /* Create hypotheses for the tags on the word, depending on word kind */
    switch (word->kind)
    {
        case Known:
            node->base = copy_from_dict(new, word->d);
            break;

        case Unknown:
            node->base = copy_unknown(new, word->text, word->d);
            break;

        case Special:
            node->base = copy_from_dict(new, word->d);
            break;

        case PhraseStart:
        case PhraseEnd:
            if (InOpt(lancpars))
            {
                /* Create a single hypothesis from the correct tag */
                node->base = create_link(NULL, FALSE,
                    create_lex_hyp(word->ctag, 1.0, NULL, new),
                    NULL, NULL);
            }
            else
            {
                fprintf(stderr,
                    "Consistency error: PhraseStart/PhraseEnd seen without lancpars\n");
                get_out();
            }
            break;
    }

    /* Check for success */
    if (node->base == NULL)
        error_exit1("No hypotheses on word %s\n", WordText(word));

    /* Copy base list to start and end lists */
    node->start = copy_hyps(node->base, node, node);
    node->end   = copy_links(node->start);

    return new;
}

/*-----------------------------------------------------------------------------
    free_base_hyps

    Free the list of base hypotheses on a node.
-----------------------------------------------------------------------------*/

static void free_base_hyps(Node n)
{
    Link l, next;

    for (l = n->base ; l != NULL ; l = next)
    {
        next = l->next;
        free(l->u.hyp);
        free(l);
    }
}

/*-----------------------------------------------------------------------------
    free_links

    Free links. If "full" is not set, then just frees the links. Otherwise,
    frees the links, and the scored hypothesis structures they point to.
-----------------------------------------------------------------------------*/

void free_links(Link l, BOOL full)
{
    Link next;

    for ( ; l ; l = next)
    {
        next = l->next;
        if (full)
            free(l->u.shyp);
        free(l);
    }
}

/*-----------------------------------------------------------------------------
    free_nodes

    Free a range of nodes, and the links attached to it. Scans the hyp list
    for phrases and recurses on them. Also free the hypothesis list for this
    node.
-----------------------------------------------------------------------------*/

void free_nodes(Node from, Node to)
{
    Node n, next;

    for (n = from ; n != NULL ; n = next)
    {
#ifdef Phrasal
        Link l;

        /* Free phrases and active edges */
        for (l = n->start ; l != NULL ; l = l->next)
        {
            Hyp hyp = l->u.shyp->hyp;

            if (hyp->type == PhraseHyp)
                /* Free the subsumed nodes */
                free_nodes(hyp->p.phrase.start, hyp->p.phrase.end);
#ifdef Use_Parser
            else if (hyp->type == ActiveHyp)
                parser_free_hyp(hyp);
#endif
        }
#endif

        /* Free the start and end links */
        free_links(n->start, FALSE);
        free_links(n->end,   TRUE);

        /* Free the base hypotheses assigned at this node */
        free_base_hyps(n);

#ifdef Use_FSM
        fsm_free_node(n);
#endif
        next = Succ(n,to);
        free(n);
    }
}

/*-----------------------------------------------------------------------------
    free_stack

    Free a range of entries from the stack. The flag indicates if the word is
    to be freed as well. Includes freeing the associated nodes.
-----------------------------------------------------------------------------*/

void free_stack(Lexeme from, Lexeme to, BOOL do_word)
{
    Node n, next, end = to->node;

    for (n = from->node ; n != NULL ; n = next)
    {
        Lexeme lp = n->lex;
        next = Succ(n, end);

        /* Free the word if so told */
        if (do_word) free_word(lp->word);

        /* Free the node (and hypotheses) */
        free_nodes(lp->node, lp->node);

        /* Free the lexeme itself */
        free(lp);
    }
}

/*-----------------------------------------------------------------------------
    is_start_ambig

    Tests if a node is ambiguous in the hypotheses starting here.
-----------------------------------------------------------------------------*/

BOOL is_start_ambig(Node n)
{
#ifdef Use_FSM
    if (n->fsm_state != NULL) return TRUE;
#endif
    return (n->start != NULL && n->start->next != NULL);
}

/*-----------------------------------------------------------------------------
    is_end_ambig

    Tests if a node is ambiguous in the hypotheses ending here.
-----------------------------------------------------------------------------*/

BOOL is_end_ambig(Node n)
{
#ifdef Use_FSM
    if (n->fsm_state != NULL) return TRUE;
#endif
    return (n->end != NULL && n->end->next != NULL);
}
